/* Name: asp162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2009-12-16
 * Original Project: USBaspLoader
 * Original Author: Christian Starkjohann
 * Original Creation Date: 2007-12-08
 * Tabsize: 8
 *
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * Copyright: (c) 2009 Thomas Fischl <tfischl@gmx.de> (original USBasp)
 * Copyright: (c) 2009 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: GNU GPL v2 (see License.txt)
 * Original Revision: Id: main.c 730 2009-03-20 09:05:11Z cs 
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <usbconfig.h>
#include <usb162.h>
#ifdef EP_TX
#include <cdc162.h>
#endif
#include <isp.h>
#include <clock.h>

//#include <avr/boot.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#include "aspconfig.h"


//#define NO_INTERRPT
#ifdef NO_INTERRPT
#define __cli()
#define __sei()
#else
#define __cli()	__asm__ volatile ("cli" )
#define __sei()	__asm__ volatile ("sei" )
#endif

/* ------------------------------------------------------------------------ */

/* Request constants used by USBasp */
#define USBASP_FUNC_CONNECT         1
#define USBASP_FUNC_DISCONNECT      2
#define USBASP_FUNC_TRANSMIT        3
#define USBASP_FUNC_READFLASH       4
#define USBASP_FUNC_ENABLEPROG      5
#define USBASP_FUNC_WRITEFLASH      6
#define USBASP_FUNC_READEEPROM      7
#define USBASP_FUNC_WRITEEEPROM     8
#define USBASP_FUNC_SETLONGADDRESS  9
#define USBASP_FUNC_SETISPSCK       10

/* Block mode flags */
#define PROG_BLOCKFLAG_FIRST    1
#define PROG_BLOCKFLAG_LAST     2

/* Block mode data size */
#define USBASP_READBLOCKSIZE   200
#define USBASP_WRITEBLOCKSIZE  200

/* ------------------------------------------------------------------------ */

/* device compatibility: */
#ifndef GICR    /* ATMega*8 don't have GICR, use MCUCR instead */
#   define GICR     MCUCR
#endif

/* ------------------------------------------------------------------------ */
#   define CURRENT_ADDRESS  currentAddress.l
#   define addr_t           uint32_t

typedef union longConverter{
    addr_t  l;
    uint16_t  w[sizeof(addr_t)/2];
    uint8_t   b[sizeof(addr_t)];
}longConverter_t;

static longConverter_t  currentAddress; /* in bytes */
static uint8_t bytesRemaining;
static uint8_t isLastPage;
static uint8_t currentRequest;
static uint8_t prog_sck = USBASP_ISP_SCK_AUTO;
//static uint8_t prog_blockflags;
static int16_t prog_pagemask;

/* ------------------------------------------------------------------------ */

uint8_t asp_setup(uint8_t *in, uint8_t *out)
{
    uint8_t  len = 0;
    uint8_t  request = in[USB_RQ_REQ]; 
    uint8_t  value_l = in[USB_RQ_VALUE_L];
    uint8_t  value_h = in[USB_RQ_VALUE_H];
    uint8_t  rval = 0;

    busy_led_on();

    if(request == USBASP_FUNC_TRANSMIT){
	out[0] = ispTransmit(value_l); /* offset 2 */
	out[1] = ispTransmit(value_h); /* offset 3 */
	out[2] = ispTransmit(in[USB_RQ_INDEX_L]); /* offset 4 */
	out[3] = ispTransmit(in[USB_RQ_INDEX_H]); /* offset 5 */
        len = 4;
    }else if(request == USBASP_FUNC_ENABLEPROG){
        out[0] = ispEnterProgrammingMode();
        len = 1;
    }else if(request == USBASP_FUNC_SETISPSCK){
	/* set sck option */
	prog_sck = value_l; /* offset 2 */
        out[0] = 0;
        len = 1;
    }else if(request == USBASP_FUNC_CONNECT){
	/* set SCK speed */
	ispSetSCKOption(prog_sck);
	connect_led_on();
	ispConnect();
    }else if(request == USBASP_FUNC_DISCONNECT){
	ispDisconnect();
	connect_led_off();
    }else if(request == USBASP_FUNC_SETLONGADDRESS){
	/* set new address */
        currentAddress.b[2] = in[USB_RQ_INDEX_L];
        currentAddress.b[3] = in[USB_RQ_INDEX_H];
    }else if(request > USBASP_FUNC_SETISPSCK) {
    }else {
        currentAddress.b[0] = value_l;
        currentAddress.b[1] = value_h;
	len = in[USB_RQ_LENGTH_L];
	if(request == USBASP_FUNC_READEEPROM || request == USBASP_FUNC_READFLASH) {
    	    for(rval = 0; rval < len; rval++) {
	    	if (request == USBASP_FUNC_READEEPROM) {
        	    *out++ = ispReadEEPROM(CURRENT_ADDRESS);
		} else {
            	    *out++ = ispReadFlash(CURRENT_ADDRESS);
		}
        	CURRENT_ADDRESS++;
	    }
        } else { /* if(request == USBASP_FUNC_WRITEEEPROM || request == USBASP_FUNC_WRITEFLASH) */
		isLastPage = in[USB_RQ_INDEX_H] & PROG_BLOCKFLAG_LAST;
		prog_pagemask = in[USB_RQ_INDEX_L]; /* offset 4 */
		//prog_blockflags = in[USB_RQ_INDEX_H] & 0x0F; /* offset 5 */
		prog_pagemask += (((unsigned int) in[USB_RQ_INDEX_H] & 0xF0) << 4);
		prog_pagemask = prog_pagemask -1;
		currentRequest = request;
		bytesRemaining = len;
		len = 0; // NO ERROR @ OUT
        }
    }
    busy_led_off();
    return len;
}

uint8_t asp_out(uint8_t *data, uint8_t len)
{
    uint8_t isLast;

    busy_led_on();
    if(len > bytesRemaining)
        len = bytesRemaining;
    bytesRemaining -= len;
    isLast = (bytesRemaining == 0);
    while (len--) {
        if(currentRequest == USBASP_FUNC_WRITEEEPROM){
	    ispWriteEEPROM(CURRENT_ADDRESS, *data++);
        } else if(prog_pagemask <= 0) {
	    ispWriteFlash(CURRENT_ADDRESS, *data++, 1);
        } else { // USBASP_FUNC_WRITEFLASH w paged
	    ispWriteFlash(CURRENT_ADDRESS, *data++, 0);
	    if (((currentAddress.w[0] & prog_pagemask) == prog_pagemask)
		|| (isLast && isLastPage && (len == 0))) {
	        ispFlushPage(CURRENT_ADDRESS, 0);
	    }
        }
	CURRENT_ADDRESS ++;
    }
    busy_led_off();
    return isLast?(uint8_t)255:0; 
}

/* ------------------------------------------------------------------------ */

static inline void asp_init() {
	uint8_t i = 20;
	while(--i) {
		_delay_ms(1);
	}
#if 0
	usbDeviceDisconnect();
	while (--i) {
		_delay_ms(1);
	}
#endif
	usb_vendor_setup = asp_setup;
	usb_user_out = asp_out;
	usbInit();
	sei();
}

int main(void)
{ 
#ifdef F_SCALER
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = F_SCALER;
#endif

    // bit_set(REGCR, REGDIS); // disable internal regulator

    /* initialize  */
    wdt_disable();      /* main app may have enabled watchdog */
    cli();
    clockInit();
    busy_led_off();

    asp_init();

    //connect_led_on();
    for (;;) {
        usbPoll();
    }
    //usbDeviceDisconnect();
    return 0;
}
