/*
 *	Copyright (c)2009	KAWAKAMI Yukio
 *	2009.10.9	Version 0.1.1
 *	2009.4.19	Version 0.1.0
 *	Copyright (c)2010	Koji Suzuki	
 *	2010.12.7	ported to angel_loader/usbasp2
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "isp.h"
#include "pdi.h"
#include "tpi.h"
#include "clock.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#define TX_ERROR		(-1)
#define RX_NODATA		(-1)
#define RX_ERROR_PARITY		0x100
#define RX_ERROR_FRAMING	0x200
#define RX_ERROR_OVERRUN	0x400

#include "aspconfig.h"

#define SEND_BUFSIZ		32
#define RECV_BUFSIZ		32

#define RECV_TIMEOUT		2	// 4-8ms
#define SEND_TIMEOUT		2	// 4-8ms

#define NVMBUSY_TIMEOUT		5	// 16-20ms
#define CHIPERASE_TIMEOUT	10	// 36-40ms

static inline void pditpi_dir_init() {
	bit_set(ISP2_OUT, ISP2_DIR);
	bit_set(ISP2_DDR, ISP2_DIR);
}

static inline void pditpi_dir_tx() {
	bit_clear(ISP2_OUT, ISP2_DIR);
}

static inline void pditpi_dir_rx() {
	bit_set(ISP2_OUT, ISP2_DIR);
}

static inline int8_t pditpi_dir_is_tx() {
	return bit_is_clear(ISP2_OUT, ISP2_DIR);
}

static inline void tpi_reset(uchar is_assert) {
	if (is_assert) {
		bit_set(ISP2_DDR, ISP2_RST); // L 
	} else {
		bit_clear(ISP2_DDR, ISP2_RST); // HI-Z
	}
}

static int8_t is_tpi;

// USART Rx buffer
static volatile uint16_t	Rbuff[RECV_BUFSIZ];
static volatile uint8_t	Rremain = 0;
static uint8_t			Rin = 0;
static uint8_t			Rout = 0;

static volatile uint8_t	timeout1_timer;
static volatile uint8_t	timeout2_timer;

static uint8_t pdi_pgmsk;
static uint8_t pdi_len;
static uint8_t blk_len;
static uint32_t pdi_addr;

static const uint8_t key_val[8] = { 0xff,0x88,0xd8,0xcd,0x45,0xab,0x89,0x12 };

static uint16_t tpi_addr;

struct pdi_space_conf {
   uint8_t read;
   uint8_t clear_buff;
   uint8_t load_buff;
   uint8_t erase_page;
   uint8_t write_page;
};

static const struct pdi_space_conf pdi_space_conf[] = {
  /* space0: ALL_FLASH */
  { PDI_NVM_READ_NVM // PDI_NVM_READ_FLASH
  , PDI_NVM_CLEAR_FLASH_BUFF
  , PDI_NVM_FILL_FLASH_BUFF
  , PDI_NVM_ERASE_FLASH_PAGE
  , PDI_NVM_WRITE_FLASH_PAGE } , 

  /* space1: APP_FLASH */
  { PDI_NVM_READ_FLASH
  , PDI_NVM_CLEAR_FLASH_BUFF
  , PDI_NVM_FILL_FLASH_BUFF
  , PDI_NVM_ERASE_APP_PAGE
  , PDI_NVM_WRITE_APP_PAGE } , 

  /* space2: BOOT_FLASH */
  { PDI_NVM_READ_FLASH
  , PDI_NVM_CLEAR_FLASH_BUFF
  , PDI_NVM_FILL_FLASH_BUFF
  , PDI_NVM_ERASE_BOOT_PAGE
  , PDI_NVM_WRITE_BOOT_PAGE } , 

  /* space3: EEPROM */
  { PDI_NVM_READ_EEPROM
  , PDI_NVM_CLEAR_EEPROM_BUFF
  , PDI_NVM_FILL_EEPROM_BUFF
  , PDI_NVM_ERASE_EEPROM_PAGE
  , PDI_NVM_WRITE_EEPROM_PAGE } , 

  /* space4: FUSE */
  { PDI_NVM_READ_FUSE
  , PDI_NVM_NOP
  , PDI_NVM_WRITE_FUSE
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

  /* space5: LOCKBITS */
  { PDI_NVM_READ_FUSE
  , PDI_NVM_NOP
  , PDI_NVM_WRITE_LOCKBIT
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

  /* space6: USERSIG */
  { PDI_NVM_READ_USERSIG
  , PDI_NVM_CLEAR_FLASH_BUFF
  , PDI_NVM_FILL_FLASH_BUFF
  , PDI_NVM_ERASE_USERSIG
  , PDI_NVM_WRITE_USERSIG } , 

  /* space7: CALIBRATION */
  { PDI_NVM_READ_CALIBRATION
  , PDI_NVM_NOP
  , PDI_NVM_NOP
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

};

static uint8_t pdi_space;
#define PDI_NVM_READ		(pdi_space_conf[pdi_space].read)
#define PDI_NVM_CLEAR_BUFF	(pdi_space_conf[pdi_space].clear_buff)
#define PDI_NVM_FILL_BUFF	(pdi_space_conf[pdi_space].load_buff)
#define PDI_NVM_ERASE_PAGE	(pdi_space_conf[pdi_space].erase_page)
#define PDI_NVM_WRITE_PAGE	(pdi_space_conf[pdi_space].write_page)

/* memo : 
 PDI_NVM_READ_NVM can read from all (?)
*/
static void pditpi_flush() {
	wdt_reset();
	if (bit_is_set(UCSR1B,TXEN1)) {
		timeout1_timer = SEND_TIMEOUT;
		while (timeout1_timer) {
			if (bit_is_set(UCSR1A,TXC1)) {
				bit_set(UCSR1A,TXC1);
				break;
			}
		}
	}
	cli();
	Rremain = 0;
	Rin = Rout = 0;
	sei();

	UCSR1B = (1<<RXEN1)|(1<<RXCIE1);
	pditpi_dir_rx();
	bit_clear(ISP2_DDR,ISP2_MOSI);
	bit_clear(ISP2_OUT,ISP2_MOSI); // TX = HI-Z
}

static uchar tpipdi_sck;
static uchar tpipdi_delay16; // (>= USBASP_ISP_SCK_250)
 
static void pditpiSetSCKOption(uchar option) {
	if ((option == USBASP_ISP_SCK_AUTO) || ( option > USBASP_ISP_SCK_8000)) 
		option = USBASP_ISP_SCK_250;
	if (option < USBASP_ISP_SCK_250)
		option = USBASP_ISP_SCK_250;
#if F_CPU == 8000000UL
	if ( option > USBASP_ISP_SCK_4000) 
		option = USBASP_ISP_SCK_4000;
	UBRR1 = (1 << (14 - option)) - 1;
#elif F_CPU == 16000000UL
	UBRR1 = (1 << (15 - option)) - 1;
#elif F_CPU == 12000000UL
	if ( option > USBASP_ISP_SCK_2000) 
		option = USBASP_ISP_SCK_2000;
	UBRR1 = (3 << (13 - option)) - 1;
#endif
	tpipdi_sck = option;
#if F_CPU == 8000000UL
	tpipdi_delay16 = 2 << (USBASP_ISP_SCK_8000 - option);
#elif F_CPU == 12000000UL
	tpipdi_delay16 = 3 << (USBASP_ISP_SCK_8000 - option);
#elif F_CPU == 16000000UL
	tpipdi_delay16 = 4 << (USBASP_ISP_SCK_8000 - option);
#endif
	tpipdi_delay16 = ( (tpipdi_delay16 +3) >> 2 ) +1;
}

static void pditpi_wait_xck() {
	if (tpipdi_sck < USBASP_ISP_SCK_1000) {
		uint8_t count = 16;
		while (--count) {
			while (bit_is_set(ISP2_IN, ISP2_SCK))
					;
			while (bit_is_clear(ISP2_IN, ISP2_SCK))
					;
		}
	} else {
		clockDelay(tpipdi_delay16);
	}
}

static void pditpiHWdisable() {
	pditpi_flush();

	/* timer stop */
	bit_clear(TIMSK1,OCIE1A);
	TCCR1B = 0;
	OCR1A = 0;

	UCSR1B = 0;	      //disable interrputs
	UCSR1C = 0;

	if (is_tpi) {
		tpi_reset(0);
	}

    	bit_set(ISP2_OUT,ISP2_MOSI);
	bit_clear(ISP2_DDR,ISP2_MOSI);  // TX = pull-up 
    	bit_set(ISP2_OUT,ISP2_SCK);
    	bit_clear(ISP2_DDR,ISP2_SCK);  // XCK = pull-up

	// clear buffers again
	pditpi_flush();
}

static void pditpiHWenable() {
	pditpi_dir_init();

	if (is_tpi) {
		tpi_reset(1);
		clockWait(200); // 64ms
		clockWait(200); // 64ms
	}
	bit_clear(ISP2_DDR,ISP2_MISO);
	bit_clear(ISP2_OUT,ISP2_MISO);  // RX = HI-Z
	bit_set(ISP2_OUT,ISP2_MOSI);
	bit_set(ISP2_DDR,ISP2_MOSI); // TX = H
    	bit_set(ISP2_OUT,ISP2_SCK);
    	bit_set(ISP2_DDR,ISP2_SCK);  // XCK = H

	clockWait(1); // 320us wait before enable XCK

	bit_clear(UCSR1A , U2X1); // must clear in syncronus mode
	UCSR1C = (1<<UMSEL10) | (2<<UPM10) | (1<<USBS1)
			| (3<<UCSZ10) | (1<<UCPOL1);

	//enable interrputs
	UCSR1B = (1<<RXEN1)|(1<<RXCIE1);

	pditpi_wait_xck();
	pditpi_wait_xck();

	// 250Hz timer
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(3<<CS10);  // CTC  1/64
	TCCR1C = 0;
	OCR1A =  (F_CPU/64/250 - 1) & 0xffff;
	bit_set(TIMSK1,OCIE1A);
}


ISR(TIMER1_COMPA_vect) {
	if (timeout1_timer) {
		--timeout1_timer;
	}
	if (timeout2_timer) {
		--timeout2_timer;
	}
}

static uchar pditpi_need_break;

ISR(USART1_RX_vect) {
	int16_t c;
	uint8_t flags;
	flags = UCSR1A;
	c = UDR1;
	if (bit_is_set(flags, UPE1)) {
		if (!is_tpi && (c == 0xDB || c == 0xEB)) // DELAY or EMPTY
			return;
		c |= RX_ERROR_PARITY; // BREAK or real error
	}
	if (bit_is_set(flags, FE1)) {
		c |= RX_ERROR_FRAMING; // real break ?
	}
	if (bit_is_set(flags, DOR1)) {
		c |= RX_ERROR_OVERRUN;
	}
	Rbuff[Rin++] = c;
	if (Rin >= sizeof(Rbuff)/sizeof(Rbuff[0])) {
		Rin = 0;
	}
	Rremain++;
}


static int8_t pditpi_send(uint8_t data) {
	if (bit_is_clear(UCSR1B,TXEN1)) {
		bit_set(UCSR1A, TXC1);
		bit_set(ISP2_OUT,ISP2_MOSI);
		bit_set(ISP2_DDR,ISP2_MOSI);
		pditpi_dir_tx();
		if (pditpi_need_break) {
			UCSR1B = 0;
			bit_clear(ISP2_OUT,ISP2_MOSI);
			pditpi_wait_xck();
			bit_set(ISP2_OUT,ISP2_MOSI);
			pditpi_need_break = 0;
		}
		pditpi_wait_xck();
		UCSR1B = (1<<TXEN1);
	}
	timeout1_timer = SEND_TIMEOUT;
	while (timeout1_timer) {
		if (bit_is_set(UCSR1A, UDRE1)) {
			UDR1 = data;
			return 0;
		}
		wdt_reset();
	}
	return TX_ERROR;
}

static int16_t pditpi_recv() {
	int16_t c = RX_NODATA;
	timeout1_timer = RECV_TIMEOUT;
	while (timeout1_timer) {
	    if (Rremain) {
		c = Rbuff[Rout++];
		if (Rout >= sizeof(Rbuff)/sizeof(Rbuff[0])) {
			Rout = 0;
		}
		cli();
		--Rremain;
		sei();
		return (c);
	    }
	    wdt_reset();
	}
	pditpi_need_break |= (c >> 8);
	return (c);
}

static void pditpiDisconnect() {
	/* disable hardware */
	pditpiHWdisable();
}

/* TPI operations */

static int16_t tpiReadCS(unsigned int address) {
	uint8_t inst;
	int16_t r;
	inst = address & 0xf;
	if (address & 0x8000) { /* read CS */
		inst |= 0x80; // TPI_SLDCS
	} else { /* read IO */
		inst |= (address << 1) & 0x60;	
		inst |= 0x10; // TPI_SIN
	}

	r = pditpi_send(inst);
	pditpi_flush();
	if (!r) r = pditpi_recv();
	return r;
}

static uchar tpiWriteCS(unsigned int address, uchar data) {
	uint8_t inst;
	inst = address & 0xf;
	if (address & 0x8000) { /* write CS */
		inst |= 0xc0; // TPI_SSTCS
	} else {
		inst |= (address << 1) & 0x60;
		inst |= 0x90; // TPI_SOUT
	}
	pditpi_send(inst);
	pditpi_send(data);
	return 0;
}

static uchar tpiSetPR(uint16_t addr) {
	uint8_t inst;
	inst = 0x68; // TPI_SETPR_LO
	pditpi_send(inst);
	pditpi_send(addr);

	inst = 0x69; // TPI_SETPR_HI
	pditpi_send(inst);
	pditpi_send(addr >> 8);
	return 0;
	
}

static int16_t tpiWaitNVM() {
	int16_t r = -1;
	timeout2_timer = NVMBUSY_TIMEOUT;
	while (timeout2_timer) {
		r = tpiReadCS(TPI_NVMCSR);
		if (!(r & 0xff00) && ((uchar)r & !(1<< TPI_NVMBSY))) {
			return 0;
		}
		wdt_reset();
	}
	return r;
}

static int16_t tpiReadDS(unsigned long address) {
	uint8_t inst;
	int16_t r;
	if (address != tpi_addr) {
		tpiWriteCS(TPI_NVMCMD, TPI_NVM_NOP);
		tpiSetPR(address);
		tpi_addr = address;
	}
	inst = 0x24; // TPI_SLD_POSTI
	r = pditpi_send(inst);
	pditpi_flush();
	if (!r) r = pditpi_recv();
	tpi_addr ++;
	return r;
}

static uchar tpiWriteDS(unsigned long address, uchar data, uchar pollmode) {
	uint8_t inst;
	if (address != tpi_addr) {
		if (!pollmode)
			tpiWriteCS(TPI_NVMCMD, TPI_NVM_WRITE);
		tpiSetPR(address);
		tpi_addr = address;
	}

	inst = 0x64; // TPI_SST_POSTI
	pditpi_send(inst);
	pditpi_send(data);
	if ((uint8_t)tpi_addr & 0x01) 
		tpiWaitNVM();
	tpi_addr ++;
	return 0;
}

static uchar tpiEnable() {
	int8_t i;
	uint8_t inst;
	const uint8_t *p;
	int16_t r;
	uchar count = 4;

	while (count--) {
		tpi_addr = 0xffff;

		inst = 0xe0;	// TPI_SKEY
		pditpi_send(inst);
		p = key_val;
		for (i=0; i < 8; i++) {
			pditpi_send(*p++);
		}
		i = 10;
		while (--i) {
			r = tpiReadCS(TPI_CS_TPISR);
			if (!(r & 0xff00) && ((uchar)r & (1<< TPI_CS_NVMEN))) {
				return 0;
			}
			wdt_reset();
		}
		pditpiHWdisable();
		pditpiHWenable();
	}
	return 1; /* error: device dosn't answer */
}

/* read and write 4 bytes from pditpi */
static void tpiTransmit4(uchar *out, uchar *in) {
	uint8_t tmp;
	uint8_t ispcmd = in[0];
	int16_t r = 0;
	uint16_t addr;

	addr = ((uint16_t)in[1]<<8) + in[2];

	tpi_addr = 0xffff;
        if(ispcmd == ISPCMD_READ_SIGNATURE) {
            tmp = in[2] & 3;
            r = tpiReadDS(TPI_DS_SIGNATURE + tmp);
        } else if(ispcmd == ISPCMD_WRITE_EEPROM) {
	    r = tpiWriteCS(addr,in[3]);
        } else if(ispcmd == ISPCMD_READ_EEPROM) {
	    r = tpiReadCS(addr);
        } else if(ispcmd == ISPCMD_CHK_BUSY) {
	    r = tpiWaitNVM();
        } else if(ispcmd == ISPCMD_ISP_CONF && in[2] == 0) {
	    if(in[1] == ISPCONF_CHIP_ERASE) {
		tpiWriteCS(TPI_NVMCMD, TPI_NVM_CHIP_ERASE);
		tpiWriteDS(TPI_DS_FLASH + 1, 0, 1); // HI-byte required
	    }
	}
        out[2] = (r >> 8);
        out[3] = r;
}

/* PDI operations */

static inline int16_t pdiReadNVM(uint16_t addr) {
	uint8_t inst;
	int16_t r;
	inst = PDI_LDS41;

	r = pditpi_send(inst);
	if (!r) r = pditpi_send(addr);
	if (!r) r = pditpi_send(addr>>8);
	if (!r) r = pditpi_send(0x00);
	if (!r) r = pditpi_send(0x01);
	pditpi_flush();
	if (!r) r = pditpi_recv();
	return r;
}

static inline uchar pdiWriteNVM(uint16_t addr, uchar data) {
	uint8_t inst;
	inst = PDI_STS41;
	pditpi_send(inst);
	pditpi_send(addr);
	pditpi_send(addr>>8);
	pditpi_send(0x00);
	pditpi_send(0x01);
	pditpi_send(data);
	return 0;
}

static int16_t pdiReadCS(uint16_t address) {
	uint8_t inst;
	int16_t r;
	if (!(address & 0x8000)) return pdiReadNVM(address);
	inst = address & 0xf;
	inst |= PDI_LDCS;

	r = pditpi_send(inst);
	pditpi_flush();
	if (!r) r = pditpi_recv();
	return r;
}

static uchar pdiWriteCS(uint16_t address, uchar data) {
	uint8_t inst;
	if (!(address & 0x8000)) return pdiWriteNVM(address,data);
	inst = address & 0xf;
	inst |= PDI_STCS;
	pditpi_send(inst);
	pditpi_send(data);
	return 0;
}

static uchar pdiSetPR(uint32_t addr) {
	uint8_t inst;
	inst = PDI_SETPR3;
	pditpi_send(inst);
	pditpi_send(addr);
	pditpi_send(addr >> 8);
	pditpi_send(addr >> 16);
	return 0;
}

static int16_t pdiWaitNVM() {
	int16_t r = -1;
	timeout2_timer = NVMBUSY_TIMEOUT;
	while (timeout2_timer) {
		r = pdiReadCS(PDI_NVM_STATUS);
		if (!(r & 0xff00) &&
		    !((uchar)r & ((1<< PDI_NVM_BUSY)|(1<< PDI_NVM_FBUSY)))) {
			return 0;
		}
		wdt_reset();
	}
	return r; // error
}


static void pdiStartDS(uchar len, int16_t pgmsk) {
	pdi_len = len;
	pdi_pgmsk = pgmsk;
	blk_len = 0;
}

static int16_t pdiReadDS(unsigned long address) {
	uint8_t inst;
	int16_t r;
	if (!blk_len && (address != pdi_addr)) {
		pdiWriteCS(PDI_NVM_CMD, PDI_NVM_READ);
		pdiSetPR(address);
		pdi_addr = address;
	}
	if (!blk_len && pdi_len) {
		blk_len = pdi_len;
		if (blk_len > RECV_BUFSIZ) blk_len = RECV_BUFSIZ;
		pdi_len -= blk_len;
		if (blk_len) {
			inst = PDI_REPEAT1;
			pditpi_send(inst);
			pditpi_send(blk_len -1);

			inst = PDI_LD1_POSTI;
			pditpi_send(inst);
			pditpi_flush();
		}
	}
	if (blk_len) {
		blk_len--;
	} else {
		inst = 0x24; // PDI_LD1_POSTI
		pditpi_send(inst);
		pditpi_flush();
	}
	r = pditpi_recv();
	pdi_addr ++;
	return r;
}

static uchar pdiWriteDS(unsigned long address, uchar data, uchar pollmode) {
	uint8_t inst;
	if (!blk_len && (address != pdi_addr)) {
		pdi_addr = address;
		pdiSetPR(pdi_addr);
	}
	if (!blk_len && pdi_len) {
		blk_len = pdi_len;
		if (blk_len > SEND_BUFSIZ) blk_len = SEND_BUFSIZ;
		if (pdi_pgmsk > 0) {
			uint16_t tmp = (uint16_t)pdi_addr & pdi_pgmsk;
			tmp = pdi_pgmsk + 1 - tmp;
			if (blk_len > tmp) blk_len = tmp;
		}
		pdi_len -= blk_len;
		if (blk_len) {
			inst = 0xa0; // PDI_REPEAT1
			pditpi_send(inst);
			pditpi_send(blk_len -1);

			inst = 0x64; // PDI_ST1_POSTI
			pditpi_send(inst);
		}
	}
	if (blk_len) {
		blk_len--;
	} else {
		inst = 0x64; // PDI_ST1_POSTI
		pditpi_send(inst);
	}
	pditpi_send(data);
	pdi_addr ++;
	return 0;
}

static uchar pdiEnable() {
	int8_t i;
	uint8_t inst;
	const uint8_t *p;
	int16_t r;
	uchar count = 4;

	while (count--) {
		pdi_len = 0;
		blk_len = 0;
		pdi_addr = 0xffffffff;


		pdiWriteCS(PDI_CS_CTRL, PDI_CS_IDLE_8); // idle bits  = 2 + 8
		pdiReadCS(PDI_CS_CTRL);

		pdiWriteCS(PDI_CS_RESET, PDI_CS_RESET_KEY);
		i = 10;
		while (--i && (pdiReadCS(PDI_CS_RESET) != 0x00)) {
			wdt_reset();
		}
		inst = 0xe0;	// PDI_KEY
		pditpi_send(inst);
		p = key_val;
		for (i=0; i < 8; i++) {
			pditpi_send(*p++);
		}
		i = 10;
		while (--i) {
			r = pdiReadCS(PDI_CS_STATUS);
			if (r == (1<< PDI_CS_NVMEN)) {
				return 0;
			}
			wdt_reset();
		}
		pditpiHWdisable();
		pditpiHWenable();
	}
	return 1; /* error: device dosn't answer */
}

static uchar pdiOpenPage(unsigned long address, uchar pollvalue) {
	pdiWaitNVM();
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_CLEAR_BUFF);
	pdiWriteCS(PDI_NVM_CTRLA, (1<<PDI_NVM_CMDEX));
	pdiWaitNVM();
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_FILL_BUFF);
	return 0;
}

static uchar pdiClosePage(unsigned long address, uchar pollvalue) {
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_ERASE_PAGE);
	pdiWriteDS(pdi_addr -2 , 0, 0);
	pdiWaitNVM();
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_WRITE_PAGE);
	pdiWriteDS(pdi_addr , 0, 0);
	pdiWaitNVM();
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_NOP);
	pditpi_flush();
	return 0;
}

/* read and write 4 bytes from pditpi */
static void pdiTransmit4(uchar *out, uchar *in) {
	int16_t r = 0;
	uint16_t addr;
	int8_t tmp;
	uint8_t ispcmd = in[0];
	addr = ((uint16_t)in[1]<<8) + in[2];

	pdi_len = 0;
	blk_len = 0;
	pdi_addr = 0xffffffff;

        if (ispcmd == ISPCMD_READ_SIGNATURE) {
            tmp = in[2] & 3;
	    //pdi_space = PDI_SPACE_FLASH;
            r = pdiReadCS(PDI_MCU_DEVID0 + tmp);
        } else if(ispcmd == ISPCMD_WRITE_EEPROM) {
	    r = pdiWriteCS(addr,in[3]);
        } else if(ispcmd == ISPCMD_READ_EEPROM) {
	    r = pdiReadCS(addr);
        } else if(ispcmd == ISPCMD_CHK_BUSY) {
	    r = pdiWaitNVM();
        } else if(ispcmd == ISPCMD_ISP_CONF && in[2] == 0) {
	    if(in[1] == ISPCONF_CHIP_ERASE) {
		pdiWriteCS(PDI_NVM_CMD, PDI_NVM_CHIP_ERASE);
		pdiWriteCS(PDI_NVM_CTRLA, (1<<PDI_NVM_CMDEX));
	    } else if(in[1] == ISPCONF_SET_SPACE) {
		pdi_space = 0;
		if (in[3] < 8) pdi_space = in[3];
	    } else if(in[1] == ISPCONF_RESET) {
		pdiWriteCS(PDI_CS_RESET, in[3]) ;
		timeout2_timer = NVMBUSY_TIMEOUT;
		while (timeout2_timer && (pdiReadCS(PDI_CS_RESET) != 0x00)) {
			wdt_reset();
		}
	    }
        } else {
            /* ignore all others, return default value == 0 */
        }
	pditpi_flush();
        out[2] = (r >> 8);
        out[3] = r;
}

void pditpiConnect(struct pgm_ops *ops, uchar sck_opt, uchar __is_tpi) {
	pditpiSetSCKOption(sck_opt);
	is_tpi = __is_tpi;
	pditpiHWenable();

	ops->pgm_disconnect   = pditpiDisconnect;
	if (is_tpi) {
		ops->pgm_transmit     = tpiTransmit4;
		ops->pgm_enable       = tpiEnable;
		ops->pgm_read_space0  = (uchar (*)(uint16_t))tpiReadCS;
		ops->pgm_write_space0 = tpiWriteCS;
		ops->pgm_read_space1  = (uchar (*)(uint32_t))tpiReadDS;
		ops->pgm_write_space1 = tpiWriteDS;
	} else {
		ops->pgm_transmit     = pdiTransmit4;
		ops->pgm_enable       = pdiEnable;
		ops->pgm_read_space0  = (uchar (*)(uint16_t))pdiReadCS;
		ops->pgm_write_space0 = pdiWriteCS;
		ops->pgm_read_space1  = (uchar (*)(uint32_t))pdiReadDS;
		ops->pgm_write_space1 = pdiWriteDS;
		ops->pgm_blk_start    = pdiStartDS;
		ops->pgm_open_page    = pdiOpenPage;
		ops->pgm_close_page   = pdiClosePage;
	}
}
